import setuptools

setuptools.setup(
    name='pybtex-apa-style',
    version='1.3',
    author='Naeka',
    author_email='contact@naeka.fr',
    description='Pybtex APA-like style',
    url='https://github.com/naeka/pybtex-apa-style',
    py_modules=['formatting.apa', 'labels.apa', 'names.firstlast'],
    entry_points={
        'pybtex.style.formatting': 'apa = formatting.apa:APAStyle',
        'pybtex.style.labels': 'apa = labels.apa:LabelStyle',
        'pybtex.style.names': 'firstlast = names.firstlast:NameStyle',
    },
    classifiers=(
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Text Processing :: Markup',
        'Topic :: Utilities',
    ),
)



from pathlib import Path

import pytest
from click.testing import CliRunner

from teachbooks.cli import main as commands

SERVE_DIR = Path(".")
WORK_DIR = Path("./.teachbooks")
PATH_BOOKS = Path(__file__).parent.joinpath("books")


@pytest.fixture()
def cli():
    """Provides a click.testing CliRunner object for invoking CLI commands."""
    runner = CliRunner()
    yield runner
    del runner


def test_serve_stop(cli: CliRunner):
    book = PATH_BOOKS.joinpath("01")
    build_result = cli.invoke(commands.build, book.as_posix())
    assert build_result.exit_code == 0, build_result.output

    serve_result = cli.invoke(commands.serve)
    assert serve_result.exit_code == 0, serve_result.output

    stop_result = cli.invoke(commands.stop)
    assert stop_result.exit_code == 0, stop_result.output

    # Don't clean book, as we can continue using build directory


def test_path(cli: CliRunner):
    """Test basic usage: serve path <path>"""
    book = PATH_BOOKS.joinpath("01")

    path_result = cli.invoke(commands.serve, ["path", book.as_posix()])
    assert path_result.exit_code == 0, path_result.output

    serve_result = cli.invoke(commands.serve)
    assert serve_result.exit_code == 0, serve_result.output

    stop_result = cli.invoke(commands.stop)
    assert stop_result.exit_code == 0, stop_result.output


# To add:
# - check output to see that directory moved
# - confirm that server is running at expected path


def test_path_unavailable(cli: CliRunner):
    """Test basic usage: serve path <path>"""
    path_unavailable = PATH_BOOKS
    path_available = PATH_BOOKS.joinpath("01")

    path_result = cli.invoke(commands.serve, ["path", path_unavailable.as_posix()])
    assert path_result.exit_code == 0, path_result.output

    serve_result = cli.invoke(commands.serve, ["path", path_available.as_posix()])
    assert serve_result.exit_code == 0, serve_result.output

    stop_result = cli.invoke(commands.stop)
    assert stop_result.exit_code == 0, stop_result.output



from pathlib import Path

from teachbooks.external_content import GIT_PATH
from teachbooks.external_content.process_toc import read_cloned_repos

BOOK_ROOT = Path(__file__).parent / "testbook" / "book"
CLONED_REPOS = read_cloned_repos(BOOK_ROOT / GIT_PATH / "cloned_repos.txt")



"""Add headers to external .md, .rst and .ipynb files."""

import json
from pathlib import Path


def format_header(cfg: dict, base_url: str, version: str) -> str:
    """Generate the admonition header based on the user's config."""
    admonition = (
        "```{" + cfg["attribution_color"] + "} Attribution\n"
        ":class: attribution\n"
        f"This page originates from {base_url},"
        f" version: {version}\n"
        "```\n"
    )
    if cfg["attribution_location"] == "top":
        return admonition

    return f"````{{margin}}\n{admonition}````\n"


def add_origin_notes(
    repo: Path, cfg: dict[str, str], base_url: str, version: str
) -> None:
    """Add a note denoting the origin of a certain file.

    Args:
        repo: Path to the repository git cloned by the enternal-content routine.
        cfg: TeachBooks configuration.
        base_url: Base URL of the file's repository.
        version: Name of the version (tag, branch or commit hash).
    """
    header = format_header(cfg, base_url, version)
    add_header_admonitions(repo, header)


def add_header_admonitions(repo: Path, header: str):
    """Add header to a file.

    Args:
        repo: Path to the git repo cloned by the external content routine.
        header: Preformatted admonition header
    """
    md_files = repo.glob("**/*.md")
    for md_file in md_files:
        prepend(md_file, header)

    nb_files = repo.glob("**/*.ipynb")
    for nb_file in nb_files:
        add_nb_admonition(nb_file, header)

    rst_files = repo.glob("**/*.rst")
    for rst_file in rst_files:
        add_rst_admonition(rst_file, header)


def prepend(file: Path, text: str):
    """Prepend string `text` to plaintext file `file`."""
    original_content = file.read_text(encoding="utf-8")
    file.write_text(text + original_content, encoding="utf-8")


def add_rst_admonition(file: Path, header: str):
    """Add an admonition top of reST file.

    To do this we make use of the `include` directive and write the admonition
    as a separate markdown file which will be parsed by myst.
    """
    admon_file = file.parent / f"_ad-{file.stem}.md"
    admon_file.write_text(header, encoding="utf-8")

    admonition = (
        f".. include:: {admon_file.name}\n    :parser: myst_parser.docutils_\n\n"
    )
    prepend(file, admonition)


def add_nb_admonition(file: Path, header: str):
    """Add an admonition containing `text` to the top of notebook `file."""
    notebook = json.loads(file.read_text(encoding="utf-8"))

    # Split over newlines, but add newline char back in at end of lines.
    source = header.split("\n")
    source = [line + "\n" for line in source]

    admonition_cell = {
        "cell_type": "markdown",
        "metadata": {},
        "source": source,
    }

    notebook["cells"] = [admonition_cell] + notebook["cells"]

    with file.open("w") as f:
        json.dump(notebook, f)



"""Main CLI module."""

import shutil
from pathlib import Path

import click

from teachbooks.external_content.process_toc import (
    chmod_git_files,
    process_external_toc_entries,
)


@click.group()
@click.version_option()
def main():
    """TeachBooks command line tools."""
    pass


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.argument("path-source", type=click.Path(exists=True, file_okay=True))
@click.option("--release", is_flag=True, help="Build book with release strategy")
@click.option(
    "--publish", is_flag=True, help="--public is deprecated. Use --release instead."
)
@click.option("--process-only", is_flag=True, help="Only pre-process content")
@click.pass_context
def build(ctx, path_source: str, publish: bool, release: bool, process_only: bool):
    """Pre-process book contents and run Jupyter Book build command."""
    from jupyter_book.cli.main import build as jupyter_book_build

    from teachbooks.release import copy_ext, make_release

    if publish:
        click.secho(
            "Warning: --publish is deprecated, use --release instead",
            fg="yellow",
            err=True,
        )

    strategy = "release" if release or publish else "draft"
    echo_info(f"running build with strategy '{strategy}'")
    path_src_folder = Path(path_source).absolute()
    if release or publish:
        path_conf, path_toc = make_release(path_src_folder)
        path_ext = path_src_folder / "_ext"
        if path_ext.exists():
            mg = "copying _ext/ directory to support APA in release [TEMPORARY FEATURE]"
            echo_info(click.style(mg, fg="yellow"))
            copy_ext(path_src_folder)
    else:
        path_conf = path_src_folder / "_config.yml"
        path_toc = path_src_folder / "_toc.yml"

    # Parse out external git entries from ToC
    path_toc = process_external_toc_entries(
        path_toc, path_toc.with_stem("_toc_with_local_paths"), book_root=path_src_folder
    )

    if not process_only:
        all_args = [str(path_src_folder)]
        if path_conf and path_conf.exists():
            all_args.extend(["--config", str(path_conf)])
        if path_toc and path_toc.exists():
            all_args.extend(["--toc", str(path_toc)])
        if ctx.args:
            all_args.extend(ctx.args)

        jupyter_book_build.main(args=all_args, standalone_mode=False)

        # Calculate and report build size
        build_dir = path_src_folder / "_build"
        total_size = sum(f.stat().st_size for f in build_dir.rglob("*") if f.is_file())
        size_mb = total_size / (1024 * 1024)
        echo_info(f"Build complete. Total size: {size_mb:.2f}MB")

        check_server()


@main.command()
@click.argument("path-source", type=click.Path(exists=True, file_okay=True))
@click.option("--external", is_flag=True, help="Empty _git/ directory.")
def clean(path_source, external: bool = False):
    """Stop teachbooks server and run Jupyter Book clean command."""
    from jupyter_book.cli.main import clean as jupyter_book_clean

    from teachbooks.serve import Server, ServerError

    workdir = Path(path_source) / ".teachbooks" / "server"

    # Check if a server is running and stop it if so
    try:
        server = Server.load(workdir)
        if server.is_running:
            echo_info("Stopping running server before cleaning...")
            server.stop()
            echo_info("Server stopped.")
    except ServerError:
        echo_info("No running server found.")

    # Clean external content
    gitdir = Path(path_source) / "_git"
    if external:
        if gitdir.exists():
            echo_info(f"Cleaning cloned git repositories in {gitdir}")
            shutil.rmtree(gitdir.absolute(), onerror=chmod_git_files)
        else:
            echo_info(f"No _git directory found at {gitdir}")
    else:
        if gitdir.exists():
            echo_info(
                "Skipping external content cleaning. Use --external"
                " to remove cloned git repositories."
            )

    # Now proceed with cleaning
    echo_info(f"Cleaning build artifacts in {path_source}...")
    jupyter_book_clean.main([str(path_source)])
    echo_info("Clean complete.")


@main.group(invoke_without_command=True)
# @click.argument("path-source", type=click.Path(exists=True, file_okay=True))
# @click.option("--test", is_flag=True, help="Build book with release strategy")
@click.option("-v", "--verbose", count=True)
@click.pass_context
def serve(ctx, verbose):
    """Start a web server to interact with the book locally.

    If serve dir path not provided, default is `./book/_build/html`.
    Checks to see if server is already running.
    """
    from teachbooks import BOOK_SERVE_DIR, SERVER_WORK_DIR
    from teachbooks.serve import Server

    if verbose > 0:
        echo_info("serve command invoked.")

    if ctx.invoked_subcommand is None:
        try:
            server = Server.load(Path(SERVER_WORK_DIR))
            if verbose > 0:
                echo_info(" server already exists")

            stdout_summary(server)
        except:  # noqa: E722 TODO: handle more gracefully.
            if verbose > 0:
                echo_info("no server found, creating a new one.")

            serve_dir = Path(BOOK_SERVE_DIR)

            if not serve_dir.exists():
                echo_info(
                    click.style("default directory not found: ", fg="yellow")
                    + f"{serve_dir}"
                )

                serve_dir = Path(".")
                print(
                    "            "
                    + click.style("serving current directory: ", fg="yellow")
                    + f"{serve_dir}"
                )
                print(
                    "            "
                    + click.style(
                        "specify a directory with: 'teachbooks serve path <path>'",
                        fg="yellow",
                    )
                )

            serve_path(serve_dir, verbose)


@serve.command()
@click.option("-v", "--verbose", count=True)
@click.argument("path-source", type=click.Path(exists=True, file_okay=True))
def path(path_source, verbose, no_build=False):
    """Specify relative path of directory to serve."""
    from teachbooks import BUILD_DIR, SERVER_WORK_DIR
    from teachbooks.serve import Server

    dir_with_build = Path(path_source).joinpath(BUILD_DIR)
    if dir_with_build.exists():
        serve_dir = dir_with_build
        echo_info("_build/html available and appended to path.")
    else:
        serve_dir = Path(path_source)

    echo_info(f"attempting to serve this directory: {serve_dir}")
    try:
        server = Server.load(Path(SERVER_WORK_DIR))
        if server.servedir == serve_dir:
            print("            " + "  ---> already serving this directory.")
            stdout_summary(server)
        else:
            print("            " + "  ---> already serving a different directory.")
            print("            " + "  ---> updating server directory...")
            server.stop()
            serve_path(serve_dir, verbose)
    except:  # noqa: E722 TODO: handle more gracefully.
        if verbose > 0:
            echo_info("no server found, creating a new one.")
        serve_path(serve_dir, verbose)


@serve.command()
def stop():
    """Stop the webserver."""
    from teachbooks import SERVER_WORK_DIR
    from teachbooks.serve import Server

    try:
        server = Server.load(Path(SERVER_WORK_DIR))
        server.stop()
        echo_info("server stopped.")
    except:  # noqa: E722 TODO: handle more gracefully.
        echo_info("no server found.")


def serve_path(servedir: str, verbose: int) -> None:
    """Start web server with specific path and verbosity."""
    from teachbooks import SERVER_WORK_DIR
    from teachbooks.serve import Server

    server = Server(
        servedir=Path(servedir), workdir=Path(SERVER_WORK_DIR), stdout=verbose
    )
    server.start(options=["--all"])
    stdout_summary(server)


def check_server():
    """Check if webserver is running and print status."""
    from teachbooks import SERVER_WORK_DIR
    from teachbooks.serve import Server

    try:
        server = Server.load(Path(SERVER_WORK_DIR))
        stdout_summary(server)
    except:  # noqa: E722 TODO: handle more gracefully
        echo_info("Use `teachbooks serve` to start a local server.")


def echo_info(message: str) -> None:
    """Wrapper for writing to stdout."""
    prefix = click.style("TeachBooks: ", fg="cyan", bold=True)
    click.echo(prefix + message)


def stdout_summary(server) -> None:
    """Print summary of server status."""
    echo_info(click.style(f"server running on: {server.url}", fg="green"))
    print(
        "            "
        + click.style(f"serving directory: {server.servedir}", fg="green")
    )
    print("            " + "To stop server, run: 'teachbooks serve stop'")



from teachbooks.external_content.config import check_plugins, find_config, find_plugins
from tests.test_external_content import BOOK_ROOT, CLONED_REPOS


def test_find_config_book():
    assert find_config(BOOK_ROOT.parent) == BOOK_ROOT / "_config.yml"


def test_find_config_repo():
    cfg_path = find_config(
        BOOK_ROOT / "_git/github.com_EXCITED-CO2_workshop_tutorial/v1.0.0"
    )
    expected_path = (
        BOOK_ROOT
        / "_git/github.com_EXCITED-CO2_workshop_tutorial/v1.0.0/book/_config.yml"
    )
    assert cfg_path == expected_path


def test_find_config_missing(tmp_path, capsys):
    result = find_config(tmp_path)
    assert result is None
    err = capsys.readouterr()
    assert "no config detected" in err.err


def test_find_multiple_configs(tmp_path, capsys):
    for name in ("a", "b"):
        directory = tmp_path / name
        directory.mkdir()
        (directory / "_config.yml").touch()
    result = find_config(tmp_path)
    assert result is None
    err = capsys.readouterr()
    assert "more than one config detected" in err.err


def test_find_plugins():
    expected = {
        "sphinx_image_inverter",
        "download_link_replacer",
        "sphinx.ext.extlinks",
        "sphinx.ext.imgconverter",
        "jupyterbook_patches",
    }

    assert find_plugins(BOOK_ROOT / "_config.yml")[0] == expected


def test_find_myst_expensions():
    expected = (
        {"sphinx.ext.imgconverter", "jupyterbook_patches", "download_link_replacer"},
        {"amsmath", "dollarmath", "linkify"},
    )
    cfg_path = (
        BOOK_ROOT / "_git/github.com_TeachBooks_HOS-workbook/main/book/_config.yml"
    )
    assert find_plugins(cfg_path) == expected


def test_check_plugins(capsys):
    result = check_plugins(BOOK_ROOT / "_config.yml", CLONED_REPOS)
    expected = ({"fake_test_plugin"}, {"amsmath", "dollarmath", "linkify"})
    assert result == expected

    err = capsys.readouterr()
    assert "sphinx plugins do not match" in err.err
    assert "MYST extensions do not match" in err.err


def test_check_plugins_no_conflict(capsys):
    no_conflict_repos = [repo for repo in CLONED_REPOS if "EXCITED" in str(repo)]
    result = check_plugins(BOOK_ROOT / "_config.yml", no_conflict_repos)
    expected = (set(), set())
    assert result == expected

    err = capsys.readouterr()
    assert err.err == ""



"""Teachbooks release workflow."""

import os
import re
from pathlib import Path


def make_release(sourcedir: Path) -> tuple[Path, Path]:
    """Pre-process files in a Jupyter Book directory."""
    # Make hidden directory that will contain the cleaned-up files
    workdir = sourcedir.joinpath(".teachbooks", "release")

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    for file in ["_config.yml", "_toc.yml"]:
        clean_yaml(sourcedir.joinpath(file), workdir.joinpath(file))

    return workdir.joinpath("_config.yml"), workdir.joinpath("_toc.yml")


def copy_ext(sourcedir: Path) -> None:
    """Copy _ext/ to support APA in release [TEMPORARY]."""
    # Make hidden directory that will contain the cleaned-up files
    ext_dir = sourcedir.joinpath("_ext")
    workdir = sourcedir.joinpath(".teachbooks", "release")

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    try:
        for root, dirs, files in os.walk(ext_dir):
            for _dir in dirs:
                os.makedirs(
                    workdir.joinpath(
                        "_ext", Path(root).relative_to(ext_dir).joinpath(_dir)
                    ),
                    exist_ok=True,
                )
            for file in files:
                src_file = Path(root).joinpath(file)
                dest_file = workdir.joinpath(
                    "_ext", Path(root).relative_to(ext_dir).joinpath(file)
                )
                os.makedirs(dest_file.parent, exist_ok=True)
                with open(src_file, "rb") as fsrc, open(dest_file, "wb") as fdst:
                    fdst.write(fsrc.read())
        print("Copied _ext/ directory successfully.")
    except:  # noqa: E722 (TODO: isolate specific exception, or re-raise error.)
        print("Error copying _ext/ directory.")


def clean_yaml(path_source: str | Path, path_output: str | Path) -> None:
    """Removes marked sections from a yaml file.

    A marked section can be:
        - ``# <START|END> REMOVE-FROM-PUBLISH``
        - or ``# <START|END> REMOVE-FROM-RELEASE``

    Does not require a specific indentation and can be used an
    unlimited number of times in the ``*.yml`` file. Commonly
    applied to ``_toc.yml`` and ``_config.yml`` files of a book.

    Example:
        To remove sub_page_2 and sub_page_3 from publishing::

            - file: subdirectory_1/intro_page
            sections:
            - file: subdirectory_1/sub_page_1
            # START REMOVE-FROM-PUBLISH
            - file: subdirectory_1/sub_page_2
            # END REMOVE-FROM-PUBLISH
            # START REMOVE-FROM-RELEASE
            - file: subdirectory_1/sub_page_3
            # END REMOVE-FROM-RELEASE
            - file: subdirectory_2/intro_page

    """
    with open(path_source, encoding="utf8") as f:
        yaml_source = f.read()

    # Regex to remove both PUBLISH and RELEASE tags
    re_pub_release = (
        r"# START REMOVE-FROM-(PUBLISH|RELEASE)(.|\n)*?"
        r"# END REMOVE-FROM-(PUBLISH|RELEASE)"
    )
    yaml_output = re.sub(re_pub_release, "", yaml_source)

    with open(path_output, mode="w", encoding="utf8") as f:
        f.write(yaml_output)



TEACHBOOKS_WORK_DIR = ".teachbooks/"
SERVER_WORK_DIR = TEACHBOOKS_WORK_DIR + "server/"
RELEASE_WORK_DIR = TEACHBOOKS_WORK_DIR + "release/"
BOOK_DIR = "book/"
BUILD_DIR = "_build/html/"
BOOK_SERVE_DIR = BOOK_DIR + BUILD_DIR



"""Serve a teachbook locally."""

import os
import pickle
import platform
import socket
import sys
from pathlib import Path
from subprocess import DEVNULL
from time import sleep
from typing import TypeVar

import psutil

STATUS = {"Linux": "sleeping", "Darwin": "running", "Windows": "running"}

Server_t = TypeVar("Server_t", bound="Server")


class ServerError(Exception):
    """Server exception class."""

    ...


class Server:
    """Class for managing a Python webserver in the background."""

    statefile = "state.pickle"

    def __init__(
        self,
        servedir: Path | str,
        workdir: Path | str,
        port: int | None = None,
        stdout: int | None = None,
    ) -> None:
        """Construct Server object.

        Args:
            servedir: Directory to serve.
            workdir: Directory for temporary files.
            port: (Optional) port for server, by default None. If left empty, a free
                port will be chosen automatically.
            stdout: (Optional) verbosity level, by default None (prints summary info).
                Use int={0, 1, 2, 3} with cli.
                With int=0 is silent; None is mixed, int>0.
        """
        self.stdout = stdout
        # Check if workdir already contains a statefile.
        try:
            old_instance = self.load(workdir)
            self.__dict__ = old_instance.__dict__
        except ServerError:
            self.servedir = Path(servedir)
            self.workdir = Path(workdir)
            self.port = port

            self._pid = None
            self._statepath = self.workdir / self.statefile

            if not os.path.exists(self.workdir):
                os.makedirs(self.workdir)

    def start(self, options: list[str] = None) -> bool:
        """Start server.

        Raises:
            RuntimeError: If server could not be started
        """
        if not self.servedir.is_dir():
            raise NotADirectoryError(f"Directory does not exist: {self.servedir}")

        if self.port is None:
            self.port = self._find_port()

        if self.is_running:
            if self.stdout is None or self.stdout > 0:
                print("Server already running:")
                print(f"  Serving directory: {self.servedir}")
                print(f"  Accessible at url: {self.url}")
            return
        else:
            if self.stdout is None or self.stdout > 0:
                print("Starting server:")
                print(f"  Directory: {self.servedir}")
                print(f"  Port:      {self.port}")
                print(f"  At url:    {self.url}")

            base_command = [sys.executable, "-u", "-m", "http.server", str(self.port)]
            if options:
                base_command.extend(options)

            # Print the full command for verification
            if self.stdout is None or self.stdout > 1:
                print("Starting server with this command:\n", "  ".join(base_command))

            proc = psutil.Popen(
                [sys.executable, "-u", "-m", "http.server", str(self.port)],
                cwd=self.servedir,
                stderr=DEVNULL,
                stdout=DEVNULL,
            )

            self._pid = proc.pid

            sleep(0.2)

            # Check if the subprocess is still running
            if not self.is_running:
                proc.terminate()
                msg = (
                    "Error launching the server. Perhaps a server "
                    "is already running on the selected port?"
                )
                raise RuntimeError(msg)
            else:
                self._save()

    def stop(self, options: list[str] = None) -> None:
        """Stop server and clean up."""
        try:  # noqa: SIM105 TODO: check if his rule improves this.
            psutil.Process(pid=self._pid).terminate()
        except psutil.NoSuchProcess:
            pass

        if os.path.exists(self._statepath):
            os.remove(self._statepath)

        self._pid = None

    def _save(self) -> None:
        """Save current Server object as a pickle file."""
        with open(self._statepath, "wb") as f:
            pickle.dump(self, f)

    @property
    def is_running(self) -> bool:
        """Check if the current process ID is a running webserver.

        Returns:
            True if the current process ID is running and is a webserver.
        """
        # Make sure the process exists
        try:
            proc = psutil.Process(pid=self._pid)
        except psutil.NoSuchProcess:
            return False
        isalive = proc.status() == STATUS[platform.system()]
        isserver = proc.cmdline()[1:] == ["-u", "-m", "http.server", str(self.port)]
        return isalive and isserver

    @property
    def url(self) -> str:
        """Get URL of running server.

        Returns:
            URL of the server.
        """
        return f"http://localhost:{self.port}"

    @staticmethod
    def _find_port() -> int:
        """Find open port.

        Returns:
            Port number.
        """
        # https://stackoverflow.com/a/1365284
        sock = socket.socket()
        sock.bind(("", 0))
        return sock.getsockname()[1]

    @classmethod
    def load(cls, workdir: Path | str) -> Server_t:
        """Construct a Server object from an existing pickle file.

        Args:
            workdir: Directory containing the pickle file.

        Returns:
            Server object reconstructed from pickle file.

        Raises:
            ServerError: If pickle file cannot be found.
        """
        try:
            with open(workdir / cls.statefile, "rb") as f:
                server = pickle.load(f)
        except FileNotFoundError as exc:
            raise ServerError("Server information not found.") from exc
        return server



import shutil
from pathlib import Path

import pytest
from click.testing import CliRunner

from teachbooks.cli import main as commands

WORK_DIR = Path(__file__).parent / ".teachbooks"
PATH_TESTDATA = Path(__file__).parent / "testbook"

ADMONITION_HTML = (
    '<div class="attribution admonition">'
    '<p class="admonition-title">Attribution</p>'
    "<p>This page originates from"
)


@pytest.fixture()
def cli():
    """Provides a click.testing CliRunner object for invoking CLI commands."""
    runner = CliRunner()
    yield runner
    del runner


def strip_whitespace(text: str) -> str:
    text = text.replace(" ", "")
    text = text.replace("\n", "")
    text = text.replace("\r", "")
    return text.replace("\t", "")


def test_build(cli: CliRunner, tmp_path: Path, monkeypatch):
    monkeypatch.chdir(tmp_path)  # Fix for windows on CI (otherwise relative dirs fail)
    # Copy test book to temp dir
    testdir: Path = shutil.copytree(PATH_TESTDATA, tmp_path / "test")
    bookdir = testdir / "book"

    # remove cloned_repos.txt file (to make clone routines run)
    repos = bookdir / "_git" / "cloned_repos.txt"
    repos.unlink()

    # Modify book config to be CI environment compatible
    config = bookdir / "_config.yml"
    testconfig = bookdir / "_config.test.yml"
    config.unlink()
    testconfig.rename(config)

    # Actually run tests
    build_result = cli.invoke(
        commands.build,
        bookdir.as_posix(),
        env={"NO_GIT_CLONE": ""},
    )
    assert build_result.exit_code == 0, build_result.output

    indexfile = bookdir / "_build" / "html" / "index.html"
    _gitdir = bookdir / "_git"
    notebook_page = (
        bookdir
        / "_build"
        / "html"
        / "_git"
        / "github.com_EXCITED-CO2_workshop_tutorial"
        / "v1.0.0"
        / "book"
        / "ARCO-ERA5.html"
    )
    rst_page = (
        bookdir
        / "_build"
        / "html"
        / "_git"
        / "github.com_EXCITED-CO2_workshop_tutorial"
        / "v1.0.0"
        / "book"
        / "extra_file.html"
    )
    md_page = (
        bookdir
        / "_build"
        / "html"
        / "_git"
        / "github.com_TeachBooks_HOS-workbook"
        / "main"
        / "book"
        / "intro.html"
    )

    for path in (indexfile, _gitdir, notebook_page, rst_page, md_page):
        assert path.exists()

    for page in (notebook_page, rst_page, md_page):
        page_text = page.read_text()
        page_text = strip_whitespace(page_text)
        assert strip_whitespace(ADMONITION_HTML) in page_text

    _ = cli.invoke(
        commands.clean,
        ["--external", bookdir.as_posix()],
    )
    assert not indexfile.exists()
    assert not _gitdir.exists()



"""Find and validate external content license files."""

from pathlib import Path

import click

LICENSE_MAPPING = {
    "MIT License": "MIT",
    "GNU AFFERO GENERAL PUBLIC LICENSE": "AGPL",
    "GNU GENERAL PUBLIC LICENSE": "GPL",
    "GNU LESSER GENERAL PUBLIC LICENSE": "LGPL",
    "Mozilla Public License": "Mozilla",
    "Apache License": "Apache",
    "CREATIVE COMMONS PUBLIC LICENSE": "CCPL",
    "Creative Commons Attribution-NonCommercial": "CC-BY-NC",
    "Creative Commons Attribution 4.0": "CC-BY-4.0",
    "Attribution 4.0 International": "CC-BY-4.0",
    "CC0 1.0 Universal": "CC0",
    "BSD 3-Clause License": "BSD3",
    "European Union Public Licence": "EUPL",
}


def validate_licenses(cloned_repos: list[Path], git_path: Path, error: bool):
    """Validate if the cloned repositories all have permisive licenses."""
    for repo_path in cloned_repos:
        repo_license = find_license(repo_path)
        if repo_license is None:
            msg = f"Could not find a (valid) license in repository {repo_path}"
            if error:
                raise ValueError(msg)
            else:
                click.secho(f"WARNING: {msg}", fg="yellow")


def find_license(repo_toplevel: str | Path) -> str | None:
    """Try to detect which license is being used by a repo."""
    valid_extensions = ["", ".md", ".txt", ".rst"]
    for extention in valid_extensions:
        license_file = Path(repo_toplevel) / f"LICENSE{extention}"
        if license_file.exists():
            break
    else:
        msg = (
            f"No license file found in git repository at {repo_toplevel}.\n"
            "Licenses must be named 'LICENSE' or 'LICENSE.md'."
        )
        raise FileNotFoundError(msg)

    with license_file.open("r", encoding="utf-8") as f:
        license_content = "".join(f.readlines())

    stripped_license = license_content.lower().strip(" \n\r")

    detected_license = None
    for license_text in LICENSE_MAPPING:
        if license_text.lower().strip(" \n\r") in stripped_license:
            detected_license = LICENSE_MAPPING[license_text]
            break

    return detected_license



"""Functionality to read write and compare .bib files."""

import re
from dataclasses import dataclass
from pathlib import Path

import click

from teachbooks.external_content.config import CLICK_WARNING_KWARGS

BIB_ENTRY_RE = re.compile(r"@(\w+){([\w:-]+)")


@dataclass
class BibEntry:
    """Contains all information of a bib file entry."""

    entrytype: str
    citekey: str
    content: dict[str, str]


def read_bibfile(file: Path) -> list[BibEntry]:
    """Read bib file into list of BibEntry objects.

    Args:
        file: Path to the .bib file.

    Returns:
        List of .bib file entries.
    """
    with file.open("r") as f:
        lines = f.readlines()

    entries: list[str] = []
    for line in lines:
        if len(line.strip()) > 0:
            matches = BIB_ENTRY_RE.match(line)
            if matches:
                entries.append("")
            entries[-1] += line

    bib_entries: list[BibEntry] = []
    for entry in entries:
        e = entry.strip().splitlines()
        entrytype, citekey = BIB_ENTRY_RE.findall(e.pop(0))[0]

        content = {}
        for line in e:
            i_eq = line.find("=")  # index of = sign; splits key and value
            if i_eq != -1:
                key = line[:i_eq].strip()
                val = line[i_eq + 1 :]
                leading_br = val.find("{")
                trailing_br = len(val) - val[::-1].find("}")
                content[key] = val[leading_br + 1 : trailing_br - 1]

        if len(content) > 0:
            bib_entries.append(BibEntry(entrytype, citekey, content))
        else:
            msg = f"Malformed entry ({citekey}) found in .bib file: {file}"
            click.secho(msg)

    return bib_entries


def bib_union(bibs: list[BibEntry], additional_bibs: list[BibEntry]):
    """Join two lists of .bib file entries, checking for duplicate keys.

    If the citekey exists in both lists, the title is checked. If the title
    is the same, it is assumed that the reference already exists in the main list,
    and is skipped silently. If the titles do not match, a warning is given.

    Args:
        bibs: Main list of bib entries.
        additional_bibs: List of additional bib entries you want to add to
            the main list.

    Returns:
        Joined list of .bib file entries.
    """
    bib_citekeys = set(bib.citekey for bib in bibs)
    extra_citekeys = set(bib.citekey for bib in additional_bibs)
    overlap = bib_citekeys.intersection(extra_citekeys)

    if len(overlap) == 0:  # no overlap: clean join
        return bibs + additional_bibs

    merged_bibs = bibs.copy()
    for entry in additional_bibs:
        if entry.citekey not in overlap:
            merged_bibs.append(entry)
        else:  # citekey already exists. check if title is the same
            bib_title = find(entry.citekey, bibs).content["title"]
            if bib_title == entry.content["title"]:
                pass
            else:
                msg = (
                    f"Warning: Found duplicate citekey '{entry.citekey}'in bibfile.\n"
                    "    However, the titles did not match.\n"
                    "    References in external content might be incorrect!"
                )
                click.secho(msg, **CLICK_WARNING_KWARGS)

    return merged_bibs


def find(citekey: str, bibs: list[BibEntry]) -> BibEntry:
    """Find bib entry by citekey."""
    key_index = [bib.citekey == citekey for bib in bibs].index(True)
    return bibs[key_index]


def write_bibfile(file: Path, bibs: list[BibEntry]) -> None:
    """Write a list of BibEntries to a new file."""
    with open(file, "w") as f:
        for bib in bibs:
            content_strs = [f"  {key} = {{{val}}}" for key, val in bib.content.items()]
            entry = (
                f"@{bib.entrytype}{{{bib.citekey},\n"
                + ",\n".join(content_strs)
                + ",\n}\n\n"
            )
            f.write(entry)


def merge_bibs(bibfile: Path, repos: list[Path]) -> list[BibEntry]:
    """Merge the book's .bib file with all external bibs.

    Will return an empty list if no reference.bib files are present anywhere.
    """
    book_bib = read_bibfile(bibfile) if bibfile.exists() else []

    for repo in repos:
        repo_bibfile = find_bibfile(repo)
        if repo_bibfile is not None:
            extra_bibs = read_bibfile(repo_bibfile)
            book_bib = bib_union(book_bib, extra_bibs)

    return book_bib


def find_bibfile(repo: Path) -> Path | None:
    """Try to find the references.bib file in the usual locations."""
    expected_location = repo / "book" / "references.bib"
    if expected_location.exists():
        return expected_location

    # main dir might not be named "book"
    possible_bibs = list(repo.glob("*/references.bib"))
    if len(possible_bibs) == 0:  # bib file might be in subdir
        possible_bibs = list(repo.glob("*/*/references.bib"))
    if len(possible_bibs) == 1:
        return possible_bibs[0]
    if len(possible_bibs) > 1:
        msg = (
            "Warning: more than one references.bib file detected in external\n"
            "     content repo, could not merge references into main file."
            f"    Please check {repo}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    else:
        msg = (
            "Warning: no references.bib file detected in external content.\n"
            f"    Please check {repo}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    return None



# Many book-related tests follow that set up in
# github.com/jupyter-book/jupyter-book/
from pathlib import Path

import pytest
from click.testing import CliRunner

from teachbooks.cli import main as commands

WORK_DIR = Path("./.teachbooks")
PATH_BOOKS = Path(__file__).parent / "books"


@pytest.fixture()
def cli():
    """Provides a click.testing CliRunner object for invoking CLI commands."""
    runner = CliRunner()
    yield runner
    del runner


def test_build(cli: CliRunner):
    book = PATH_BOOKS.joinpath("01")
    build_result = cli.invoke(commands.build, book.as_posix())
    assert build_result.exit_code == 0, build_result.output
    html = book.joinpath("_build", "html")
    assert html.joinpath("index.html").exists()
    _ = cli.invoke(commands.clean, book.as_posix())
    assert not html.joinpath("index.html").exists()


def test_build_release(cli: CliRunner):
    book = PATH_BOOKS.joinpath("01")
    build_result = cli.invoke(commands.build, ["--release", book.as_posix()])
    assert build_result.exit_code == 0, build_result.output
    html = book.joinpath("_build", "html")
    assert html.joinpath("index.html").exists()
    _ = cli.invoke(commands.clean, book.as_posix())
    assert not html.joinpath("index.html").exists()


def test_build_publish(cli: CliRunner):
    book = PATH_BOOKS.joinpath("01")
    build_result = cli.invoke(commands.build, ["--publish", book.as_posix()])
    assert build_result.exit_code == 0, build_result.output
    html = book.joinpath("_build", "html")
    assert html.joinpath("index.html").exists()
    _ = cli.invoke(commands.clean, book.as_posix())
    assert not html.joinpath("index.html").exists()


def test_build_release_apa(cli: CliRunner):
    """Confirm _ext is copied into .teachbooks dir.

    Note that this does not test proper build, which
    must be done manually in a GitHub repo using the
    deploy-book-workflow with BRANCHES_TO_PREPROCESS.
    """
    book = PATH_BOOKS / "02"
    build_result = cli.invoke(commands.build, ["--release", book.as_posix()])
    assert build_result.exit_code == 0, build_result.output

    html = book / "_build" / "html"
    assert (html / "index.html").exists()

    ext_dir = book / ".teachbooks" / "release" / "_ext"
    assert (ext_dir / "apastyle.py").exists()
    assert (ext_dir / "bracket_citation_style.py").exists()
    assert (ext_dir / "pybtexapastyle").exists()
    assert (ext_dir / "pybtexapastyle" / "setup.py").exists()

    _ = cli.invoke(commands.clean, book.as_posix())
    assert not (html / "index.html").exists()

    # build 2x to make sure there aren't write conflicts within _ext
    build_result = cli.invoke(commands.build, ["--release", book.as_posix()])
    assert build_result.exit_code == 0, build_result.output
    assert (html / "index.html").exists()

    assert (ext_dir / "apastyle.py").exists()
    assert (ext_dir / "bracket_citation_style.py").exists()
    assert (ext_dir / "pybtexapastyle").exists()
    assert (ext_dir / "pybtexapastyle" / "setup.py").exists()


def test_build_external_content(cli: CliRunner):
    book = PATH_BOOKS.joinpath("03", "book")
    build_result = cli.invoke(commands.build, book.as_posix())
    assert build_result.exit_code == 0, build_result.output
    assert (book / "_git").exists()
    html = book.joinpath("_build", "html")
    assert html.joinpath("index.html").exists()

    _ = cli.invoke(commands.clean, ["--external", book.as_posix()])
    assert not html.joinpath("index.html").exists()
    assert not book.joinpath("_git").exists()


def test_build_release_external_content(cli: CliRunner):
    book = PATH_BOOKS.joinpath("03", "book")
    build_result = cli.invoke(commands.build, ["--release", book.as_posix()])
    assert build_result.exit_code == 0, build_result.output
    assert book.joinpath("_git").exists()
    html = book.joinpath("_build", "html")
    assert html.joinpath("index.html").exists()

    _ = cli.invoke(commands.clean, ["--external", book.as_posix()])
    assert not html.joinpath("index.html").exists()
    assert not book.joinpath("_git").exists()



"""Parse external content config files."""

from pathlib import Path
from typing import Any

import click

from teachbooks.external_content.utils import load_yaml_file

CLICK_WARNING_KWARGS: dict[str, Any] = {"fg": "yellow", "err": True}

DEFAULT_CONFIG = {
    "attribution_color": "admonition",  # should refer to a color defined in _config.
    "attribution_location": "top",  # Either 'top' or 'margin'
}


def get_teachbooks_config(config_file: Path) -> dict[str, str]:
    """Load teachbooks entries from _config.yml.

    Args:
        config_file: Path to _config.yml file

    Returns:
        Loaded configuration.
    """
    config = load_yaml_file(config_file)

    cfg = DEFAULT_CONFIG
    if "teachbooks" in config:
        if "attribution_color" in config["teachbooks"]:
            cfg["attribution_color"] = config["teachbooks"]["attribution_color"]
        if "attribution_location" in config["teachbooks"]:
            loc = config["teachbooks"]["attribution_location"]
            if loc not in ["top", "margin"]:
                msg = (
                    "Incorrect config location configuration. "
                    "Should be `top` or `margin`"
                )
                raise ValueError(msg)
            cfg["attribution_location"] = loc
    return cfg


def check_plugins(
    config_file: Path, git_repos: list[Path]
) -> tuple[set[str], set[str]]:
    """Check external configs for missing plugins/myst extensions and return them.

    The user will be warned if any plugins or extensions are missing/not the same
    version pin.

    Args:
        config_file: Path to the config yaml file.
        git_repos: List of cloned git repositories.

    Returns:
        The missing plugins and myst extensions.
    """
    plugins, myst_extensions = find_plugins(config_file)

    missing_plugins: set[str] = set()
    missing_extensions: set[str] = set()

    for repo in git_repos:
        repo_config = find_config(repo)
        if repo_config is not None:
            repo_plugins, repo_myst_ext = find_plugins(repo_config)
            unmatched_plugins = repo_plugins - plugins
            unmatched_extensions = repo_myst_ext - myst_extensions

            missing_plugins = missing_plugins | unmatched_plugins
            missing_extensions = missing_extensions | unmatched_extensions

            if len(unmatched_plugins) > 0:
                unmatched_str = "".join(
                    [f"       {req}\n" for req in unmatched_plugins]
                )
                click.secho(
                    "Warning: sphinx plugins do not match between external content "
                    "and your book\n"
                    f"    Repository: {repo}\n"
                    "    Missing or non-matching plugins:\n"
                    f"{unmatched_str}"
                    "Content from this book might not display correctly.",
                    **CLICK_WARNING_KWARGS,
                )

            if len(unmatched_extensions) > 0:
                unmatched_str = "".join(
                    [f"       {req}\n" for req in unmatched_extensions]
                )
                click.secho(
                    "Warning: MYST extensions do not match between external content "
                    "and your book\n"
                    f"    Repository: {repo}\n"
                    "    Missing or non-matching extensions:\n"
                    f"{unmatched_str}"
                    "Content from this book might not display correctly.",
                    **CLICK_WARNING_KWARGS,
                )
    return missing_plugins, missing_extensions


def find_config(repo: Path) -> Path | None:
    """Find path to config file.

    Args:
        repo: Path to repository.

    Returns:
        Path to config file if found, otherwise None.
    """
    std_config = repo / "book" / "_config.yml"
    if std_config.exists():
        return std_config

    # book name might be different. Only search one directory level deep:
    detected_configs = list(repo.glob("*/_config.yml"))
    if len(detected_configs) == 1:
        return detected_configs[0]
    if len(detected_configs) > 1:
        msg = (
            "Warning: more than one config detected in external content repo,\n"
            "    could not validate config plugins/extensions."
            f"    Please check {repo}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    elif len(detected_configs) == 0:
        msg = (
            "Warning: no config detected in external content repo,\n"
            "    could not validate config plugins/extensions."
            f"    Please check {repo}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    return None


def find_plugins(config_file: Path) -> tuple[set[str], set[str]]:
    """Collect all plugins (sphinx, myst) from config file."""
    sphinx_plugins = get_sphinx_plugins(config_file)
    myst_extensions = get_myst_extensions(config_file)
    return sphinx_plugins, myst_extensions


def get_sphinx_plugins(config_file: Path) -> set[str]:
    """Find the sphinx plugins in the config file.

    Args:
        config_file: Path to the config yaml file.

    Returns:
        Sphinx plugin names as a set.
    """
    config = load_yaml_file(config_file)

    sphinx_plugins = set()
    if config.get("sphinx") is None:
        msg = (
            "Warning: Missing 'sphinx' entry in config,\n"
            f"    Please check {config_file}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    elif not isinstance(config.get("sphinx"), dict):
        msg = (
            "Malformed 'sphinx' entry in config (expected sub-entries),\n"
            f"    Please check {config_file}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    elif config.get("sphinx").get("extra_extensions") is None:
        msg = (
            "Malformed 'sphinx' entry in config (expected 'extra_extensions' entry),\n"
            f"    Please check {config_file}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    else:
        sphinx_plugins = set(config.get("sphinx").get("extra_extensions"))
    return sphinx_plugins


def get_myst_extensions(config_file: Path) -> set[str]:
    """Find the myst extensions in the config file.

    Args:
        config_file: Path to the config yaml file.

    Returns:
        myst extension names as a set.
    """
    config = load_yaml_file(config_file)

    myst_extensions = set()
    if config.get("parse") is None:
        pass  # "parse" entry is not required.
    elif not isinstance(config.get("parse"), dict):
        msg = (
            "Malformed 'parse' entry in config (expected sub-entries),\n"
            f"    Please check {config_file}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    elif config.get("parse").get("myst_enable_extensions") is None:
        msg = (
            "Malformed 'parse' entry in config (expected 'myst_enable_extensions'"
            f" entry),\n    Please check {config_file}"
        )
        click.secho(msg, **CLICK_WARNING_KWARGS)
    else:
        myst_extensions = set(config.get("parse").get("myst_enable_extensions"))
    return myst_extensions
